#include <assert.h>
#include "hdnum.hh"   

using namespace hdnum; // Namen ohne hdnum:: verwenden

template<class M>
class Shooting
{
public:
  /** \brief export size_type */
  typedef typename M::size_type size_type;

  /** \brief export time_type */
  typedef typename M::time_type time_type;

  /** \brief export number_type */
  typedef typename M::number_type numbertype;

public:

  Shooting (const M& model1_, const M& model2_, const M& model3_, 
	    numbertype t_a_, numbertype t_b_,
	    int n_, DenseMatrix<numbertype> b_a_, 
	    DenseMatrix<numbertype> b_b_, Vector<numbertype> g_)
    : model1(model1_), model2(model2_), model3(model3_),
      rksolver1(model1), rksolver2(model2), rksolver3(model3),
      t(t_a), t_a(t_a_), t_b(t_b_), n(n_),
      u(n+1,model1.size()), g(model1.size())
  {
    u = 0;
    b_a = b_a_;
    b_b = b_b_;
    g = g_;
    numbertype dt = (t_b-t_a)/n;
    rksolver1.set_dt(dt); //solver for y0
    rksolver2.set_dt(dt); //solver for Y0
    rksolver3.set_dt(dt); //solver for Y1
  }

  //! do one step
  void solve ()
  { 
    DenseMatrix<numbertype> solutions1(n+1,2); //store y0
    DenseMatrix<numbertype> solutions2(n+1,2); //store Y0
    DenseMatrix<numbertype> solutions3(n+1,2); //store Y1
    // store inital values for y0,Y0,Y1
    for (int j=0; j<=1; j++){
      solutions1[0][j] = (rksolver1.get_state())[j];
      solutions2[0][j] = (rksolver2.get_state())[j];
      solutions3[0][j] = (rksolver3.get_state())[j];  
    }

    // store values for y0,Y0,Y1 for each timestep
    for (int i=1; i<=n; i++){
      // ***add Code here***
    }

    // compute Q^h
    DenseMatrix<numbertype> Q(2,2); 
    Q = b_a;
    DenseMatrix<numbertype> temp(2,2,0);
    DenseMatrix<numbertype> YN(2,2,0); // compute Y_N
    YN[0][0] = solutions2[n][0];
    YN[1][0] = solutions2[n][1];
    YN[0][1] = solutions3[n][0];
    YN[1][1] = solutions3[n][1];
    // ***add Code here***

    // compute right hand side b = g-b_b*y0,N
    Vector<numbertype> b(2);
    Vector<numbertype> y0N(2);
    y0N[0] = solutions1[n][0];
    y0N[1] = solutions1[n][1];
    // ***add Code here***
  
    // solve Qs = b
    Vector<numbertype> s(2);
    Vector<numbertype> y(2);
    Array<std::size_t> p(2);
    Array<std::size_t> q(2);
    row_equilibrate(Q,y);
    s = 0;
    lr_fullpivot(Q,p,q);
    apply_equilibrate(y,b);
    permute_forward(p,b);
    solveL(Q,b,b);
    solveR(Q,s,b);
    permute_backward(q,s);

    // compute u_n
    for(int i=0; i<=n; i++){
      // ***add Code here***
    } 
  }

  //! write out u
  void output (Vector<numbertype>& times_, Vector<Vector<numbertype> >& states_)
  {
    Vector<numbertype> times;           // store time values here
    Vector<Vector<numbertype> > states; // store states here
    for(int i=0; i<=n; i++){
      numbertype dt = (t_b-t_a)/n;
      times.push_back(i*dt);
      Vector<numbertype> temp(2);
      temp[0] = u[i][0]; temp[1] = u[i][1]; 
      states.push_back(temp);
    }
    times_ = times;
    states_ = states;
  }

//! get current state
  const Vector<numbertype>& get_state () const
  {
    return u[0];
  }

  //! get current time
  time_type get_time () const
  {
    return t;
  }

  //! get dt used in last step (i.e. to compute current state)
  time_type get_dt () const
  {
    return (t_a-t_b)/n;
  }

private:
  const M& model1, model2, model3;                    
  RungeKutta4<M> rksolver1, rksolver2, rksolver3;
  numbertype t, t_a, t_b;
  int n;
  DenseMatrix<numbertype> u;
  DenseMatrix<numbertype> b_a;
  DenseMatrix<numbertype> b_b;
  Vector<numbertype> g;
};
